// MIT License
//
// Copyright (c) 2025 Cocotec Limited
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
// SPDX-License-Identifier: MIT

#include "coco/runtime.h"

// This is defined to 'inline' when this .cc file is #included in runtime.h
#ifndef COCO_DETAIL_INLINED
#define COCO_DETAIL_INLINED
#endif

#if !defined(COCO_GCC44_WORKAROUNDS) && !defined(__clang__) && defined(__GNUC__) && \
    (__GNUC__ < 4 || (__GNUC__ == 4 && __GNUC_MINOR__ < 5))
#define COCO_GCC44_WORKAROUNDS
#endif

namespace coco {
inline namespace v1 {

COCO_DETAIL_INLINED BaseComponent::BaseComponent() : coco_logger_(NULL), started_(false) {}
COCO_DETAIL_INLINED BaseComponent::~BaseComponent() = default;

COCO_DETAIL_INLINED void BaseComponent::set_name(std::string name) {
  name_ = std::move(name);
  set_names_recursively();
}

COCO_DETAIL_INLINED void BaseComponent::set_name(const std::string& base_name, const std::string& field_name) {
  name_.clear();
  name_.reserve(base_name.size() + 1 + field_name.size());
  name_.append(base_name);
  name_.push_back('.');
  name_.append(field_name);
  set_names_recursively();
}

COCO_DETAIL_INLINED void BaseComponent::set_logger_recursively(StateMachineLogger* logger) { set_logger(logger); }
COCO_DETAIL_INLINED void BaseComponent::add_observer(ComponentObserver& observer) { observers_.push_back(&observer); }
COCO_DETAIL_INLINED void BaseComponent::set_names_recursively() {}

COCO_DETAIL_INLINED void BaseComponent::send_machine_starting() {
  for (auto it = observers_.begin(), it_end = observers_.end(); it != it_end; ++it) {
    (*it)->machine_starting(*this);
  }
}

COCO_DETAIL_INLINED void BaseComponent::send_machine_terminated() {
  for (auto it = observers_.begin(), it_end = observers_.end(); it != it_end; ++it) {
    (*it)->machine_terminated(*this);
  }
}

COCO_DETAIL_INLINED BasePort::BasePort() {}
COCO_DETAIL_INLINED BasePort::~BasePort() {}
COCO_DETAIL_INLINED void BasePort::coco_start() {}

COCO_DETAIL_INLINED void BaseProvidedPort::single_threaded_subscribe(SingleThreadedSubscriber& port) {
  // If this abort is executed, this means that the component providing this port was expected to be single-threaded,
  // but is instead multi-threaded.
  (void)port;
  abort();
}

COCO_DETAIL_INLINED void BaseProvidedPort::single_threaded_unsubscribe(SingleThreadedSubscriber& port) {
  // If this abort is executed, this means that the component providing this port was expected to be single-threaded,
  // but is instead multi-threaded.
  (void)port;
  abort();
}

COCO_DETAIL_INLINED BaseMultiThreadedEncapsulatingComponent::BaseMultiThreadedEncapsulatingComponent()
    : running_machines_(0) {}
COCO_DETAIL_INLINED BaseMultiThreadedEncapsulatingComponent::~BaseMultiThreadedEncapsulatingComponent() {}
COCO_DETAIL_INLINED void BaseMultiThreadedEncapsulatingComponent::wait_for_component_termination() {
  std::unique_lock<std::mutex> lock(mutex_);
  while (!have_all_machines_terminated()) {
    all_machines_terminated_.wait(lock);
  }
}

COCO_DETAIL_INLINED void BaseMultiThreadedEncapsulatingComponent::machine_starting(const BaseComponent&) {
  std::lock_guard<std::mutex> lock(mutex_);
  if (++running_machines_ == 1) {
    send_machine_starting();
  }
}

COCO_DETAIL_INLINED void BaseMultiThreadedEncapsulatingComponent::machine_terminated(const BaseComponent&) {
  std::lock_guard<std::mutex> lock(mutex_);
  COCO_RUNTIME_ASSERT(running_machines_ > 0);
  if (--running_machines_ == 0) {
    send_machine_terminated();
    all_machines_terminated_.notify_all();
  }
}

COCO_DETAIL_INLINED BaseSingleThreadedComponent::BaseSingleThreadedComponent() : timer_executor_(NULL) {}
COCO_DETAIL_INLINED void BaseSingleThreadedComponent::set_timer_executor(SingleThreadedTimerExecutor& executor) {
  timer_executor_ = &executor;
}

COCO_DETAIL_INLINED void BaseSingleThreadedComponent::execute_timer(int) { abort(); }
COCO_DETAIL_INLINED void BaseSingleThreadedComponent::coco_drain_queue() { send_drain_queue(); }
COCO_DETAIL_INLINED void BaseSingleThreadedComponent::cancel_timer(int timer_id) {
  COCO_RUNTIME_ASSERT(timer_executor_);
  timer_executor_->cancel_timer(*this, timer_id);
}

COCO_DETAIL_INLINED void BaseSingleThreadedComponent::enable_timer(int timer_id, CocoClock::duration duration) {
  COCO_RUNTIME_ASSERT(timer_executor_);
  timer_executor_->enable_timer(*this, timer_id, duration);
}

COCO_DETAIL_INLINED void BaseSingleThreadedComponent::send_drain_queue() {
  for (auto it = single_threaded_subscribers_.begin(), it_end = single_threaded_subscribers_.end(); it != it_end;
       ++it) {
    (*it)->coco_drain_queue();
  }
}

COCO_DETAIL_INLINED void BaseSingleThreadedComponent::add_single_threaded_subscriber(
    SingleThreadedSubscriber& subscriber) {
  single_threaded_subscribers_.insert(&subscriber);
}

COCO_DETAIL_INLINED void BaseSingleThreadedComponent::remove_single_threaded_subscriber(
    SingleThreadedSubscriber& subscriber) {
  single_threaded_subscribers_.erase(&subscriber);
}

COCO_DETAIL_INLINED BaseSingleThreadedEncapsulatingComponent::BaseSingleThreadedEncapsulatingComponent()
    : running_machines_(0) {}

COCO_DETAIL_INLINED BaseSingleThreadedEncapsulatingComponent::~BaseSingleThreadedEncapsulatingComponent() = default;

COCO_DETAIL_INLINED void BaseSingleThreadedEncapsulatingComponent::all_machines_terminated() {
  send_machine_terminated();
}

COCO_DETAIL_INLINED void BaseSingleThreadedEncapsulatingComponent ::machine_starting(const BaseComponent&) {
  if (++running_machines_ == 1) {
    send_machine_starting();
  }
}

COCO_DETAIL_INLINED void BaseSingleThreadedEncapsulatingComponent ::machine_terminated(const BaseComponent&) {
  COCO_RUNTIME_ASSERT(running_machines_ > 0);
  if (--running_machines_ == 0) {
    all_machines_terminated();
  }
}

COCO_DETAIL_INLINED void BaseWrappedSingleThreadedEncapsulatingComponent::unlocked_enable_timer(
    BaseSingleThreadedComponent& component, int timer, CocoClock::duration duration) {
  auto pair =
      timer_map_.insert(std::make_pair(std::make_pair(&component, timer), static_cast<unsigned int>(timers_.size())));
  if (pair.second) {
    timers_.push_back({&component, timer});
  }
  timer_queue_.enable_timer(pair.first->second, duration);
}

COCO_DETAIL_INLINED void BaseWrappedSingleThreadedEncapsulatingComponent::unlocked_cancel_timer(
    BaseSingleThreadedComponent& component, int timer) {
  timer_queue_.cancel_timer(timer_map_.at({&component, timer}));
}

COCO_DETAIL_INLINED void BaseWrappedSingleThreadedEncapsulatingComponent::execute_timer() {
  COCO_RUNTIME_ASSERT(timer_fired());
  std::pair<BaseSingleThreadedComponent*, int> timer = timers_[timer_queue_.pop()];
  timer.first->execute_timer(timer.second);
}

COCO_DETAIL_INLINED ComponentObserver::~ComponentObserver() = default;
COCO_DETAIL_INLINED void ComponentObserver::machine_starting(const BaseComponent& component) { (void)component; }
COCO_DETAIL_INLINED void ComponentObserver::machine_terminated(const BaseComponent& component) { (void)component; }

COCO_DETAIL_INLINED Formatter::ArrayLog::ArrayLog(Formatter& formatter) : formatter_(formatter), is_first_(true) {
  formatter_.write_char('[');
}

COCO_DETAIL_INLINED Formatter::EnumLog::EnumLog(Formatter& formatter, bool positional)
    : formatter_(formatter), positional_(positional), is_first_(true) {
  formatter_.write_char(positional ? '(' : '{');
}

COCO_DETAIL_INLINED void Formatter::EnumLog::begin_field(const char* name) {
  formatter_.maybe_comma(is_first_);
  if (!positional_) {
    formatter_.write_string(name);
    formatter_.write_char('=');
  }
}

COCO_DETAIL_INLINED Formatter::StructLog::StructLog(Formatter& formatter) : formatter_(formatter), is_first_(true) {
  formatter_.write_char('{');
}

COCO_DETAIL_INLINED void Formatter::StructLog::begin_field(const char* name) {
  formatter_.maybe_comma(is_first_);
  formatter_.write_string(name);
  formatter_.write_char('=');
}

COCO_DETAIL_INLINED Formatter::~Formatter() {}

COCO_DETAIL_INLINED void Formatter::write_escaped_string(const char* string, const size_t size) {
  const char* current = string;
  const char* to_write = string;
  for (const char* end = string + size; current != end; ++current) {
    char escaped;
    switch (*current) {
      case '\0':
        escaped = '0';
        break;
      case '\n':
        escaped = 'n';
        break;
      case '\t':
        escaped = 't';
        break;
      case '\r':
        escaped = 'r';
        break;
      case '\"':
        escaped = '"';
        break;
      case '\'':
        escaped = '\'';
        break;
      case '\\':
        escaped = '\\';
        break;
      case '`': {
        escaped = '`';
        break;
      }
      default:
        continue;
    }

    size_t to_flush = static_cast<size_t>(current - to_write);
    if (to_flush > 0) {
      write_data(to_write, to_flush);
    }
    write_char('\\');
    write_char(escaped);
    to_write = current + 1;
  }
  size_t remaining = static_cast<size_t>(current - to_write);
  if (remaining > 0) {
    write_data(to_write, remaining);
  }
}

COCO_DETAIL_INLINED void Formatter::write_int(int value) {
#ifdef COCO_GCC44_WORKAROUNDS
  write_string(std::to_string(static_cast<long long int>(value)));
#else
  write_string(std::to_string(value));
#endif
}

COCO_DETAIL_INLINED void Formatter::write_uint(unsigned int value) {
#ifdef COCO_GCC44_WORKAROUNDS
  write_string(std::to_string(static_cast<long long unsigned int>(value)));
#else
  write_string(std::to_string(value));
#endif
}

COCO_DETAIL_INLINED void Formatter::write_int_padded(int value, unsigned int width) {
  std::string s =
#ifdef COCO_GCC44_WORKAROUNDS
      std::to_string(static_cast<long long unsigned int>(value));
#else
      std::to_string(value);
#endif
  if (s.size() < width) {
    std::string padding(width - s.size(), '0');
    write_string(padding);
  }
  write_string(s);
}

COCO_DETAIL_INLINED Formatter::EnumLog Formatter::log_begin_enum(const char* case_name, bool positional) {
  write_string(case_name);
  return EnumLog(*this, positional);
}

COCO_DETAIL_INLINED void Formatter::log_external_value(const char* string) {
  log_begin_external_value();
  write_escaped_string(string, strlen(string));
  log_finish_external_value();
}

COCO_DETAIL_INLINED void Formatter::log_external_value(const std::string& string) {
  log_begin_external_value();
  write_escaped_string(string.data(), string.size());
  log_finish_external_value();
}

#if __cplusplus >= 201703L
COCO_DETAIL_INLINED void Formatter::log_external_value(std::string_view string) {
  log_begin_external_value();
  write_escaped_string(string.data(), string.size());
  log_finish_external_value();
}
#endif

COCO_DETAIL_INLINED IllegalException::IllegalException(std::string component_name)
    : component_name_(std::move(component_name)), state_name_("") {}

COCO_DETAIL_INLINED IllegalException::IllegalException(std::string component_name, const char* state_name)
    : component_name_(std::move(component_name)), state_name_(state_name) {}

COCO_DETAIL_INLINED IllegalException::~IllegalException() throw() {}

COCO_DETAIL_INLINED const char* IllegalException::what() const throw() {
  return "Illegal behaviour triggered on Coco model";
}

COCO_DETAIL_INLINED SingleThreadedSubscriber::~SingleThreadedSubscriber() = default;

COCO_DETAIL_INLINED SingleThreadedTimerExecutor::SingleThreadedTimerExecutor() = default;
COCO_DETAIL_INLINED SingleThreadedTimerExecutor::~SingleThreadedTimerExecutor() = default;

COCO_DETAIL_INLINED StateMachineLogger::~StateMachineLogger() = default;
COCO_DETAIL_INLINED StateMachineLogger::EventLogger::~EventLogger() = default;
COCO_DETAIL_INLINED void StateMachineLogger::EventLogger::begin_constructed_signal() {
  // If this abort is invoked this means you are using a logger that does not support logging signal sends: either
  // remove the usage of `logValues`, downgrade to a standard below Coco 2024, or upgrade the logger to handle
  // signals.
  ::abort();
}

COCO_DETAIL_INLINED void StateMachineLogger::machine_starting(const BaseComponent&) {}
COCO_DETAIL_INLINED void StateMachineLogger::machine_started(const BaseComponent&, const char*) {}
COCO_DETAIL_INLINED void StateMachineLogger::machine_terminated(const BaseComponent&) {}
COCO_DETAIL_INLINED void StateMachineLogger::state_entry_started(const BaseComponent&, const char*) {}
COCO_DETAIL_INLINED void StateMachineLogger::state_exit_finished(const BaseComponent&, const char*) {}
COCO_DETAIL_INLINED void StateMachineLogger::event_started(const BaseComponent&, const char*, const char*,
                                                           const char*) {}
COCO_DETAIL_INLINED void StateMachineLogger::event_finished(const BaseComponent&, const char*, const char*,
                                                            const char*) {}
COCO_DETAIL_INLINED void StateMachineLogger::message(const BaseComponent&, const char*, const char*) {}
COCO_DETAIL_INLINED void StateMachineLogger::message(const BaseComponent& component, const char* current_state,
                                                     const std::string& message) {
  return this->message(component, current_state, message.c_str());
}
COCO_DETAIL_INLINED void StateMachineLogger::abort(const BaseComponent& comonent, const char*) {}
COCO_DETAIL_INLINED StateMachineLogger::EventLogger& StateMachineLogger::event_started(const BaseComponent&,
                                                                                       const char*, const char*) {
  // If this abort is invoked this means you are using a logger that does not support values whilst also setting
  // `logValues` in the `Coco.toml` file: either remove the usage of `logValues` or upgrade the logger to handle
  // values.
  ::abort();
}
COCO_DETAIL_INLINED StateMachineLogger::EventLogger& StateMachineLogger::event_finished(const BaseComponent&,
                                                                                        const char*, const char*) {
  // If this abort is invoked this means you are using a logger that does not support values whilst also setting
  // `logValues` in the `Coco.toml` file: either remove the usage of `logValues` or upgrade the logger to handle
  // values.
  ::abort();
}

COCO_DETAIL_INLINED StateMachineLogger::EventLogger& StateMachineLogger::message(const BaseComponent&) {
  // If this abort is invoked this means you are using a logger that does not support values whilst also setting
  // `logValues` in the `Coco.toml` file: either remove the usage of `logValues` or upgrade the logger to handle
  // values.
  ::abort();
}
COCO_DETAIL_INLINED StateMachineLogger::EventLogger& StateMachineLogger::send_signal(const BaseComponent&, const char*,
                                                                                     const char*) {
  // If this abort is invoked this means you are using a logger that does not support logging signal sends: either
  // remove the usage of `logValues`, downgrade to a standard below Coco 2024, or upgrade the logger to handle values.
  ::abort();
}
COCO_DETAIL_INLINED StateMachineLogger::EventLogger& StateMachineLogger::send_signal(const BaseComponent&,
                                                                                     const char*) {
  // If this abort is invoked this means you are using a logger that does not support logging signal sends: either
  // remove the usage of `logValues`, downgrade to a standard below Coco 2024, or upgrade the logger to handle
  // signals.
  ::abort();
}

COCO_DETAIL_INLINED StringFormatter::StringFormatter() = default;
COCO_DETAIL_INLINED StringFormatter::~StringFormatter() = default;
COCO_DETAIL_INLINED void StringFormatter::write_char(const char c) { buffer_.push_back(c); }
COCO_DETAIL_INLINED void StringFormatter::write_data(const char* string, size_t size) { buffer_.append(string, size); }
COCO_DETAIL_INLINED std::string StringFormatter::take() {
  std::string result = std::move(buffer_);
  buffer_ = std::string();
  return result;
}

}  // namespace v1
}  // namespace coco

#undef COCO_GCC44_WORKAROUNDS
