// MIT License
//
// Copyright (c) 2025 Cocotec Limited
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
// SPDX-License-Identifier: MIT

#include "coco/stream_logger.h"

// This is defined to 'inline' when this .cc file is #included in stream_logger.h
#ifndef COCO_DETAIL_INLINED
#define COCO_DETAIL_INLINED
#endif

namespace coco {

COCO_DETAIL_INLINED FormatterEventLogger::FormatterEventLogger(Formatter& formatter, TimestampFormat timestamp_format)
    : formatter_(formatter),
      timestamp_format_(timestamp_format),
      position_(Position::FirstParameter),
      port_(NULL),
      event_(NULL) {}

COCO_DETAIL_INLINED FormatterEventLogger& FormatterEventLogger::setup(const char* direction,
                                                                      const BaseComponent& component, const char* port,
                                                                      const char* event) {
  position_ = Position::FirstParameter;
  port_ = port;
  event_ = event;
  maybe_write_timestamp();
  formatter_.write_string(direction);
  formatter_.write_string(component.name());
  return *this;
}

COCO_DETAIL_INLINED void FormatterEventLogger::done() {
  output_prefix(false);
  if (event_ && position_ != Position::Finished) {
    formatter_.write_char('}');
  }
}

COCO_DETAIL_INLINED void FormatterEventLogger::maybe_write_timestamp() {
  switch (timestamp_format_) {
    case TimestampFormat::None:
      break;
    case TimestampFormat::ISO8601UTC:
      formatter_.write_char('[');
      format_iso8601_utc_timestamp(std::chrono::system_clock::now(), formatter_);
      formatter_.write_string("] ");
      break;
  }
}

COCO_DETAIL_INLINED void FormatterEventLogger::format_iso8601_utc_timestamp(
    std::chrono::system_clock::time_point time_point, Formatter& formatter) {
  std::time_t native_time = std::chrono::system_clock::to_time_t(time_point);
  std::tm tm{};
  gmtime(native_time, tm);

  formatter.write_int(tm.tm_year + 1900);
  formatter.write_char('-');
  formatter.write_int_padded(tm.tm_mon + 1, 2);
  formatter.write_char('-');
  formatter.write_int_padded(tm.tm_mday, 2);
  formatter.write_char('T');
  formatter.write_int_padded(tm.tm_hour, 2);
  formatter.write_char(':');
  formatter.write_int_padded(tm.tm_min, 2);
  formatter.write_char(':');
  formatter.write_int_padded(tm.tm_sec, 2);
  {
    auto epoch_time = time_point.time_since_epoch();
    auto micros = std::chrono::duration_cast<std::chrono::microseconds>(epoch_time) -
                  std::chrono::duration_cast<std::chrono::microseconds>(
                      std::chrono::duration_cast<std::chrono::seconds>(epoch_time));

    formatter.write_char('.');
    formatter.write_int_padded(static_cast<int>(micros.count()), 6);
  }
  formatter.write_char('Z');
}

COCO_DETAIL_INLINED Formatter& FormatterEventLogger::formatter() { return formatter_; }

COCO_DETAIL_INLINED void FormatterEventLogger::begin_parameter(const char* parameter_name) {
  output_prefix(true);
  formatter_.write_string(parameter_name);
  formatter_.write_char('=');
}
COCO_DETAIL_INLINED void FormatterEventLogger::end_parameter(const char*) {}

COCO_DETAIL_INLINED void FormatterEventLogger::begin_return_value() {
  output_prefix(false);
  formatter_.write_string("} -> ");
}
COCO_DETAIL_INLINED void FormatterEventLogger::end_return_value() { position_ = Position::Finished; }

COCO_DETAIL_INLINED void FormatterEventLogger::begin_current_state() { formatter_.write_string(" ("); }
COCO_DETAIL_INLINED void FormatterEventLogger::end_current_state() { formatter_.write_char(')'); }

COCO_DETAIL_INLINED void FormatterEventLogger::message(const char* message, size_t message_size) {
  output_prefix(false);
  formatter().write_data(message, message_size);
}

COCO_DETAIL_INLINED void FormatterEventLogger::begin_constructed_signal() { output_prefix(false); }

COCO_DETAIL_INLINED void FormatterEventLogger::output_prefix(bool output_comma) {
  if (position_ == Position::FirstParameter) {
    formatter_.write_string(": ");
    if (port_ && port_[0] != '\0') {
      formatter_.write_string(port_);
      formatter_.write_char('.');
    }
    if (event_) {
      formatter_.write_string(event_);
      formatter_.write_char('{');
    }
    position_ = Position::NthParameter;
  } else if (output_comma) {
    formatter_.write_char(',');
  }
}

COCO_DETAIL_INLINED FormatterMachineLogger::FormatterMachineLogger(Formatter& formatter,
                                                                   TimestampFormat timestamp_format)
    : formatter_(formatter), event_logger_(*this, timestamp_format) {}

COCO_DETAIL_INLINED void FormatterMachineLogger::machine_starting(const BaseComponent& component) {
  message(component, NULL, "starting machine");
}
COCO_DETAIL_INLINED void FormatterMachineLogger::machine_started(const BaseComponent& component,
                                                                 const char* current_state) {
  message(component, current_state, "started machine");
}
COCO_DETAIL_INLINED void FormatterMachineLogger::machine_terminated(const BaseComponent& component) {
  message(component, NULL, "terminated machine");
}
COCO_DETAIL_INLINED void FormatterMachineLogger::state_entry_started(const BaseComponent& component,
                                                                     const char* current_state) {
  message(component, current_state, "entry started");
}
COCO_DETAIL_INLINED void FormatterMachineLogger::state_exit_finished(const BaseComponent& component,
                                                                     const char* current_state) {
  message(component, current_state, "exit finished");
}
COCO_DETAIL_INLINED void FormatterMachineLogger::event_started(const BaseComponent& component,
                                                               const char* current_state, const char* port,
                                                               const char* event) {
  std::lock_guard<std::mutex> lock(mutex_);
  output_prefix(component, current_state);
  formatter_.write_string("started ");
  formatter_.write_string(port);
  formatter_.write_char('.');
  formatter_.write_string(event);
  flush_message();
}
COCO_DETAIL_INLINED void FormatterMachineLogger::event_finished(const BaseComponent& component,
                                                                const char* current_state, const char* port,
                                                                const char* event) {
  std::lock_guard<std::mutex> lock(mutex_);
  output_prefix(component, current_state);
  formatter_.write_string("finished ");
  formatter_.write_string(port);
  formatter_.write_char('.');
  formatter_.write_string(event);
  flush_message();
}
COCO_DETAIL_INLINED void FormatterMachineLogger::message(const BaseComponent& component, const char* current_state,
                                                         const char* message) {
  std::lock_guard<std::mutex> lock(mutex_);
  output_prefix(component, current_state);
  formatter_.write_string(message);
  flush_message();
}
COCO_DETAIL_INLINED void FormatterMachineLogger::abort(const BaseComponent& component, const char* current_state) {
  message(component, current_state, "abort");
}

COCO_DETAIL_INLINED StateMachineLogger::EventLogger& FormatterMachineLogger::event_started(
    const BaseComponent& component, const char* port, const char* event) {
  return event_logger_.setup("> ", component, port, event);
}
COCO_DETAIL_INLINED StateMachineLogger::EventLogger& FormatterMachineLogger::event_finished(
    const BaseComponent& component, const char* port, const char* event) {
  return event_logger_.setup("< ", component, port, event);
}
COCO_DETAIL_INLINED StateMachineLogger::EventLogger& FormatterMachineLogger::message(const BaseComponent& component) {
  return event_logger_.setup("# ", component, NULL, NULL);
}
COCO_DETAIL_INLINED StateMachineLogger::EventLogger& FormatterMachineLogger::send_signal(const BaseComponent& component,
                                                                                         const char* port,
                                                                                         const char* event) {
  return event_logger_.setup("+ ", component, port, event);
}
COCO_DETAIL_INLINED StateMachineLogger::EventLogger& FormatterMachineLogger::send_signal(const BaseComponent& component,
                                                                                         const char* port) {
  return event_logger_.setup("+ ", component, port, NULL);
}

COCO_DETAIL_INLINED FormatterMachineLogger::LockedEventLogger::LockedEventLogger(FormatterMachineLogger& logger,
                                                                                 TimestampFormat timestamp_format)
    : FormatterEventLogger(logger.formatter_, timestamp_format), logger_(logger) {}

COCO_DETAIL_INLINED void FormatterMachineLogger::output_prefix(const BaseComponent& component,
                                                               const char* current_state) {
  event_logger_.maybe_write_timestamp();
  formatter_.write_string(component.name());
  if (current_state) {
    formatter_.write_string(" (");
    formatter_.write_string(current_state);
    formatter_.write_char(')');
  }
  formatter_.write_string(": ");
}

COCO_DETAIL_INLINED FormatterMachineLogger::LockedEventLogger& FormatterMachineLogger::LockedEventLogger::setup(
    const char* direction, const BaseComponent& component, const char* port, const char* event) {
  logger_.mutex_.lock();
  FormatterEventLogger::setup(direction, component, port, event);
  return *this;
}

COCO_DETAIL_INLINED void FormatterMachineLogger::LockedEventLogger::done() {
  FormatterEventLogger::done();
  logger_.flush_message();
  logger_.mutex_.unlock();
}

COCO_DETAIL_INLINED StreamMachineLogger::StreamMachineLogger(std::ostream& out, TimestampFormat timestamp_format)
    : FormatterMachineLogger(static_cast<Formatter&>(*this), timestamp_format), out_(out) {}

COCO_DETAIL_INLINED void StreamMachineLogger::flush_message() { out_ << std::endl; }
COCO_DETAIL_INLINED void StreamMachineLogger::write_char(const char c) { out_ << c; }
COCO_DETAIL_INLINED void StreamMachineLogger::write_data(const char* string, size_t size) {
  out_.write(string, static_cast<std::streamsize>(size));
}

COCO_DETAIL_INLINED StreamMachineLogger* StreamMachineLogger::cout() {
  static StreamMachineLogger logger(std::cout);
  return &logger;
}

COCO_DETAIL_INLINED StreamMachineLogger* StreamMachineLogger::cerr() {
  static StreamMachineLogger logger(std::cerr);
  return &logger;
}

}  // namespace coco
