#ifndef COCO_STREAM_LOGGER_H_
#define COCO_STREAM_LOGGER_H_

// MIT License
//
// Copyright (c) 2019 Cocotec Limited
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.
//
// SPDX-License-Identifier: MIT

#include <iostream>
#include <mutex>

#include "coco/runtime.h"

namespace coco {

enum class TimestampFormat : std::uint8_t {
  None,
  ISO8601UTC,
};

/// An StateMachineLogger::EventLogger that is implemented in terms of a `Formatter` and outputs events in the Coco
/// standardised logging format.
///
/// Should be useful as a starting point for custom implementations of `EventLogger`.
class FormatterEventLogger : public StateMachineLogger::EventLogger {
 public:
  FormatterEventLogger(Formatter& formatter, TimestampFormat timestamp_format);

  virtual FormatterEventLogger& setup(const char* direction, const BaseComponent& component, const char* port,
                                      const char* event);

  void done() COCO_OVERRIDE;

  void maybe_write_timestamp();

  /// Prints an ISO8601-formatted timestamp to \p formatter
  static void format_iso8601_utc_timestamp(std::chrono::system_clock::time_point time_point, Formatter& formatter);

 private:
  enum class Position : std::uint8_t {
    FirstParameter,
    NthParameter,
    Finished,
  };

  Formatter& formatter_;
  const TimestampFormat timestamp_format_;
  Position position_;
  const char* port_;
  const char* event_;

  Formatter& formatter() COCO_OVERRIDE;

  void begin_parameter(const char* parameter_name) COCO_OVERRIDE;
  void end_parameter(const char*) COCO_OVERRIDE;

  void begin_return_value() COCO_OVERRIDE;
  void end_return_value() COCO_OVERRIDE;

  void begin_current_state() COCO_OVERRIDE;
  void end_current_state() COCO_OVERRIDE;

  void message(const char* message, size_t message_size) COCO_OVERRIDE;

  void begin_constructed_signal() COCO_OVERRIDE;

  void output_prefix(bool output_comma);

  static inline void gmtime(const std::time_t& time, std::tm& tm) {
#ifdef _WIN32
    ::gmtime_s(&tm, &time);
#else
    ::gmtime_r(&time, &tm);
#endif
  }
};

/// A logger that outputs to a coco::Formatter in the Coco standard logging format.
class FormatterMachineLogger : public StateMachineLogger {
 public:
  void machine_starting(const BaseComponent& component) COCO_OVERRIDE;
  void machine_started(const BaseComponent& component, const char* current_state) COCO_OVERRIDE;
  void machine_terminated(const BaseComponent& component) COCO_OVERRIDE;
  void state_entry_started(const BaseComponent& component, const char* current_state) COCO_OVERRIDE;
  void state_exit_finished(const BaseComponent& component, const char* current_state) COCO_OVERRIDE;
  void event_started(const BaseComponent& component, const char* current_state, const char* port,
                     const char* event) COCO_OVERRIDE;
  void event_finished(const BaseComponent& component, const char* current_state, const char* port,
                      const char* event) COCO_OVERRIDE;
  void message(const BaseComponent& component, const char* current_state, const char* message) COCO_OVERRIDE;
  void abort(const BaseComponent& component, const char* current_state) COCO_OVERRIDE;

  EventLogger& event_started(const BaseComponent& component, const char* port, const char* event) COCO_OVERRIDE;
  EventLogger& event_finished(const BaseComponent& component, const char* port, const char* event) COCO_OVERRIDE;
  EventLogger& message(const BaseComponent& component) COCO_OVERRIDE;
  EventLogger& send_signal(const BaseComponent& component, const char* port, const char* event) COCO_OVERRIDE;
  EventLogger& send_signal(const BaseComponent& component, const char* port) COCO_OVERRIDE;

 protected:
  explicit FormatterMachineLogger(Formatter& formatter, TimestampFormat timestamp_format);

  /// Writes a single log message.
  ///
  /// The message's contents will already have been written to `formatter_`.
  virtual void flush_message() = 0;

 private:
  class LockedEventLogger : public FormatterEventLogger {
   public:
    explicit LockedEventLogger(FormatterMachineLogger& logger, TimestampFormat timestamp_format);

    LockedEventLogger& setup(const char* direction, const BaseComponent& component, const char* port,
                             const char* event) COCO_OVERRIDE;

    void done() COCO_OVERRIDE;

   private:
    FormatterMachineLogger& logger_;
  };

  std::mutex mutex_;
  Formatter& formatter_;
  LockedEventLogger event_logger_;

  void output_prefix(const BaseComponent& component, const char* current_state);
};

/// A simple logger that outputs to a std::ostream.
///
/// This is not intended for production usage, but as an example of how a logger could be implemented.
class StreamMachineLogger : public FormatterMachineLogger, private Formatter {
 public:
  /// Returns a pointer to a shared logger that logs to stdout.
  ///
  /// This can be directly passed to BaseComponent::set_logger_recursively, for example:
  /// @code
  ///   component.set_logger_recursively(StreamMachineLogger::cout);
  /// @endcode
  static StreamMachineLogger* cout();

  /// Returns a pointer to a shared logger that logs to stderr.
  ///
  /// This can be directly passed to BaseComponent::set_logger_recursively, for example:
  /// @code
  ///   component.set_logger_recursively(StreamMachineLogger::cerr);
  /// @endcode
  static StreamMachineLogger* cerr();

  explicit StreamMachineLogger(std::ostream& out, TimestampFormat timestamp_format = TimestampFormat::None);

 private:
  std::ostream& out_;

  void flush_message() COCO_OVERRIDE;
  void write_char(const char c) COCO_OVERRIDE;
  void write_data(const char* string, size_t size) COCO_OVERRIDE;
};

}  // namespace coco

#endif
